// Example Service taken from excellent tutorial at http://www.devx.com/cplus/Article/9857

// From the command line
// Install the service using
// sc create Beacon binpath= "c:\Services\BeaconService.exe 127.0.0.1:3389 83.151.197.175:6435"
// Remove the service using
// sc delete Beacon


#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
// Windows Header Files:
#include <windows.h>
// C RunTime Header Files
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>

#include <Winsock2.h>
#include "stdio.h"
#include <shellapi.h>

#define MAX_LOADSTRING 100
#define WM_APP_NOTIFYICON       (WM_APP + 0x01)
#define SELECT_CYCLES           (3)
#define CONNECT_CYCLES          (3)
#define SELECT_TIME_SECS        (5)
#define SELECT_TIME_USECS       (0)

#define LOGFILE "C:\\Services\\BeaconLog.txt"

typedef struct _SendRec
{
  SOCKET  sReceive;
  int     nReceivePort;
  char    *szReceiveAddr;
  SOCKET  sSend;
  int     nSendPort;
  char    *szSendAddr;
} SendRec;

SERVICE_STATUS          ServiceStatus; 
SERVICE_STATUS_HANDLE   hStatus; 
 


static bool         gboListen = false;
static bool         gboConnect;
static int          gnPortOut = 3389;
static int          gnPortIn = 6435;
static char         gszAddressOut[1024] = "127.0.0.1";
static char         gszAddressIn[1024] = "127.0.0.1";
static int          gnSockTypeOut = SOCK_STREAM;
static int          gnSockTypeIn = SOCK_STREAM;
static HANDLE       ghListenHandle = 0;
static WSAData      gsWSAData;
static bool         gboClosed;
static SendRec      gsSendRecIn;
static SendRec      gsSendRecOut;
static bool         gboLogging = false;

int WriteToLog (char* str)
{
  FILE* log;

  if (gboLogging)
  {
    log = fopen(LOGFILE, "a+");
    if (log == NULL)
        return -1;
    fprintf(log, "%s\n", str);
    fclose(log);
  }
  return 0;
}



void  ServiceMain(int argc, char** argv); 
void  ControlHandler(DWORD request); 
int InitService();



void ListenProc (void);
void Quit (void);
void SendRecProc (SendRec * psSendRec);
void SetScreenText (char * szText, int nPos);
void ShowError (void);
void CloseSocket (SOCKET nSocket);



void main(int argc, char** argv) 
{ 
  int nArg;
  char szInput[1024];
  char * pcColon;

  // Read the command arguments
  nArg = 1;
  while ((nArg < argc) && (nArg < 3))
  {
    WriteToLog (argv[nArg]);

    // Is there a colon?
    pcColon = strchr (argv[nArg], ':');
    if (pcColon)
    {
      strncpy (szInput, argv[nArg], pcColon - argv[nArg]);
      szInput[pcColon - argv[nArg]] = 0;
    }
    else
    {
      strcpy (szInput, argv[nArg]);
    }

    // Store the results
    switch (nArg)
    {
      case 1:
        strcpy (gszAddressOut, szInput);
        WriteToLog (gszAddressOut);
        if (pcColon)
        {
          sscanf (pcColon + 1, "%d", & gnPortOut);
          WriteToLog (pcColon + 1);
        }
        break;
      case 2:
        strcpy (gszAddressIn, szInput);
        WriteToLog (gszAddressIn);
        if (pcColon)
        {
          sscanf (pcColon + 1, "%d", & gnPortIn);
          WriteToLog (pcColon + 1);
        }
        break;
    }
    nArg++;
  }

  SERVICE_TABLE_ENTRY ServiceTable[2];
  ServiceTable[0].lpServiceName = "Beacon";
  ServiceTable[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)ServiceMain;

  ServiceTable[1].lpServiceName = NULL;
  ServiceTable[1].lpServiceProc = NULL;
  // Start the control dispatcher thread for our service
  StartServiceCtrlDispatcher(ServiceTable);  
}

void ServiceMain(int argc, char** argv) 
{ 
  int error; 


  int nArg;
  char szInput[1024];
  char * pcColon;

  ServiceStatus.dwServiceType = 
    SERVICE_WIN32; 
  ServiceStatus.dwCurrentState = 
    SERVICE_START_PENDING; 
  ServiceStatus.dwControlsAccepted   =  
    SERVICE_ACCEPT_STOP | 
    SERVICE_ACCEPT_SHUTDOWN;
  ServiceStatus.dwWin32ExitCode = 0; 
  ServiceStatus.dwServiceSpecificExitCode = 0; 
  ServiceStatus.dwCheckPoint = 0; 
  ServiceStatus.dwWaitHint = 0; 

  hStatus = RegisterServiceCtrlHandler(
    "Beacon", 
    (LPHANDLER_FUNCTION)ControlHandler); 
  if (hStatus == (SERVICE_STATUS_HANDLE)0) 
  { 
    // Registering Control Handler failed
    return; 
  }  

  // Read the command arguments
  nArg = 1;
  while ((nArg < argc) && (nArg < 3))
  {
    WriteToLog (argv[nArg]);

    // Is there a colon?
    pcColon = strchr (argv[nArg], ':');
    if (pcColon)
    {
      strncpy (szInput, argv[nArg], pcColon - argv[nArg]);
      szInput[pcColon - argv[nArg]] = 0;
    }
    else
    {
      strcpy (szInput, argv[nArg]);
    }

    // Store the results
    switch (nArg)
    {
      case 1:
        strcpy (gszAddressOut, szInput);
        WriteToLog (gszAddressOut);
        if (pcColon)
        {
          sscanf (pcColon + 1, "%d", & gnPortOut);
          WriteToLog (pcColon + 1);
        }
        break;
      case 2:
        strcpy (gszAddressIn, szInput);
        WriteToLog (gszAddressIn);
        if (pcColon)
        {
          sscanf (pcColon + 1, "%d", & gnPortIn);
          WriteToLog (pcColon + 1);
        }
        break;
    }
    nArg++;
  }


  // Initialize Service 
  error = InitService(); 
  if (error) 
  {
    // Initialization failed
    ServiceStatus.dwCurrentState = 
        SERVICE_STOPPED; 
    ServiceStatus.dwWin32ExitCode = -1; 
    SetServiceStatus(hStatus, &ServiceStatus); 
    return; 
  } 
  // We report the running status to SCM. 
  ServiceStatus.dwCurrentState = 
    SERVICE_RUNNING; 
  SetServiceStatus (hStatus, &ServiceStatus);

  ListenProc ();

	WSACleanup();

  return; 
}

// Service initialization
int InitService() 
{ 
  int result = 0;

  if (WSAStartup (MAKEWORD (1, 1), & gsWSAData) != 0)
  {
  	MessageBox (NULL, "Error starting WinSock", "Socket error", MB_OK);
	  return -1;
  }

  gboListen = true;
  gboConnect = true;

  return (result); 
}

void ControlHandler(DWORD request) 
{ 
  switch(request) 
  { 
    case SERVICE_CONTROL_STOP: 
      gboConnect = false;
      ServiceStatus.dwWin32ExitCode = 0; 
      ServiceStatus.dwCurrentState = SERVICE_STOPPED; 
      SetServiceStatus (hStatus, &ServiceStatus);
      return; 
      break;
    case SERVICE_CONTROL_SHUTDOWN: 
      gboConnect = false;
      ServiceStatus.dwWin32ExitCode = 0; 
      ServiceStatus.dwCurrentState = SERVICE_STOPPED; 
      SetServiceStatus (hStatus, &ServiceStatus);
      return; 
      break;
    default:
      break;
  } 

  // Report current status
  SetServiceStatus (hStatus, &ServiceStatus);

  return; 
}

void ListenProc (void)
{
	SOCKET              nSockfdIn;
	SOCKET              nSockfdOut;
	struct sockaddr_in  sHostAddr;
	int                 nReturn;
	SOCKET              nAcceptfd = NULL;
	timeval             sTime;
  int                 nMemSize = 1024*10;
  HANDLE              hHandleIn;
  HANDLE              hHandleOut;
  bool                boOpt = TRUE;
  char                szInitBuffer[64];
  int                 nToSendInit;
  int                 nToSendInitPos;
  unsigned long       uArg;
	fd_set              sSelectSet;
	bool                gboConnected;
	int                 nSelectCycles;
  int                 nConnectCycles;

  if (gboListen != true)
  {
    return;
  }

  while (gboConnect)
  {
    gboListen = true;
    SetScreenText ("Initialising beacon", 0);

    nSockfdIn = socket (AF_INET, gnSockTypeIn, 0);
	  if (nSockfdIn == INVALID_SOCKET)
	  {
	    ShowError ();
      gboListen = false;
	  }
	  else
	  {
      // Set to non blocking
	    uArg = 1;
      ioctlsocket (nSockfdIn, FIONBIO , & uArg);

      setsockopt (nSockfdIn, SOL_SOCKET, SO_REUSEADDR, (char *)(& boOpt), sizeof (bool));
      SetScreenText ("Created beacon", 0);

      nSockfdOut = socket (AF_INET, gnSockTypeOut, 0);
	    if (nSockfdOut == INVALID_SOCKET)
	    {
  	    ShowError ();
  	    CloseSocket (nSockfdIn);
        gboListen = false;
	    }
	    else
	    {
        setsockopt (nSockfdOut, SOL_SOCKET, SO_REUSEADDR, (char *)(& boOpt), sizeof (bool));
        SetScreenText ("Creating bridge", 0);

	      sHostAddr.sin_family = AF_INET;
	      sHostAddr.sin_port = htons (gnPortIn);
        sHostAddr.sin_addr.s_addr = inet_addr (gszAddressIn);
	      memset (& (sHostAddr.sin_zero), '\0', 8);

        SetScreenText ("Beacon active", 0);
        gboConnected = false;
        nConnectCycles = 0;
        do
        {
  	      nReturn = connect (nSockfdIn, (struct sockaddr *)& sHostAddr, sizeof (struct sockaddr));
          // Wait for a bit
          if (nReturn == 0)
          {
            gboConnected = true;
          }

          if ((nReturn != 0) && ((WSAGetLastError () == WSAEINPROGRESS) || (WSAGetLastError () == WSAEWOULDBLOCK)))
          {
            nSelectCycles = 0;
            do
            {
						  FD_ZERO (& sSelectSet);
						  FD_SET (nSockfdIn, & sSelectSet);
						  sTime.tv_sec = SELECT_TIME_SECS;
						  sTime.tv_usec = SELECT_TIME_USECS;

              nReturn = select ((int)nSockfdIn, NULL, & sSelectSet, NULL, & sTime);

              if (nReturn != -1)
              {
  						  if (FD_ISSET (nSockfdIn, & sSelectSet))
  						  {
  						    // Connected
  						    gboConnected = true;
  						  }
  					  }
  					  else
  					  {
  					    nReturn = WSAGetLastError ();
  					    gboListen = FALSE;
  					  }
              nSelectCycles++;
            } while (((!gboConnected) && (gboListen) && (gboConnect)) && (nSelectCycles < SELECT_CYCLES));
          }
        } while (((!gboConnected) && (gboListen) && (gboConnect)) && (nSelectCycles < CONNECT_CYCLES));

        if ((gboListen) && (gboConnect) && (gboConnected))
        {
          SetScreenText ("Semi connected", 1);
        }
        else
        {
	        CloseSocket (nSockfdOut);
	        CloseSocket (nSockfdIn);
          gboListen = false;
        }

        if ((gboListen) && (gboConnect) && (gboConnected))
        {
          // Set to blocking
	        uArg = 0;
          ioctlsocket (nSockfdIn, FIONBIO , & uArg);
        }

	      while ((gboListen) && (gboConnect) && (gboConnected))
	      {
          nToSendInit = 0;
	        nReturn = recv (nSockfdIn, szInitBuffer, sizeof (szInitBuffer), 0);

          if (nReturn < 0)
          {
            ShowError ();
          }
          if (nReturn <= 0)
          {
            // Connection closed
	          CloseSocket (nSockfdOut);
	          CloseSocket (nSockfdIn);
            gboListen = false;
          }
          if (nReturn > 0)
          {
            nToSendInit = nReturn;
            SetScreenText ("Connecting bridge", 1);

	          sHostAddr.sin_family = AF_INET;
	          sHostAddr.sin_port = htons (gnPortOut);
	          sHostAddr.sin_addr.s_addr = inet_addr (gszAddressOut);
	          memset (& (sHostAddr.sin_zero), '\0', 8);
            nReturn = connect (nSockfdOut, (struct sockaddr *)& sHostAddr, sizeof (struct sockaddr));
            if (nReturn != 0)
            {
              ShowError ();
	            CloseSocket (nSockfdOut);
	            CloseSocket (nSockfdIn);
              gboListen = false;
            }
            else
            {
              SetScreenText ("Bridge active", 1);

              // Forward on the initial data
              nToSendInitPos = 0;
              while ((nToSendInitPos < nToSendInit) && (gboConnect) && (gboListen))
              {
  	            nReturn = send (nSockfdOut, szInitBuffer + nToSendInitPos, nToSendInit - nToSendInitPos, 0);
                if (nReturn > 0)
                {
                  nToSendInitPos += nReturn;
                }
                if (nReturn <= 0)
                {
                  ShowError ();
                  nToSendInit = 0;
	                CloseSocket (nSockfdOut);
	                CloseSocket (nSockfdIn);
                  gboListen = false;
                }
              }

              if (gboListen)
              {
                gboClosed = false;

                gsSendRecIn.sReceive = nSockfdIn;
                gsSendRecIn.szReceiveAddr = gszAddressIn;
                gsSendRecIn.nReceivePort = gnPortIn;
                gsSendRecIn.sSend = nSockfdOut;
                gsSendRecIn.szSendAddr = gszAddressOut;
                gsSendRecIn.nSendPort = gnPortOut;

                hHandleIn = CreateThread (NULL, 0, (LPTHREAD_START_ROUTINE) SendRecProc, (void*)(& gsSendRecIn), 0, NULL);

                gsSendRecOut.sReceive = nSockfdOut;
                gsSendRecOut.szReceiveAddr = gszAddressOut;
                gsSendRecOut.nReceivePort = gnPortOut;
                gsSendRecOut.sSend = nSockfdIn;
                gsSendRecOut.szSendAddr = gszAddressIn;
                gsSendRecOut.nSendPort = gnPortIn;

                hHandleOut = CreateThread (NULL, 0, (LPTHREAD_START_ROUTINE) SendRecProc, (void*)(& gsSendRecOut), 0, NULL);

                gboListen = false;
              }
            }
		      }
	      }
	    }
	  }
    SetScreenText ("Beacon inactive", 0);
    SetScreenText ("Bridge inactive", 1);
	}

  gboListen = false;
}

void SendRecProc (SendRec * psSendRec)
{
	int                 nStorePos;
  int                 nMemSize = 1024*10;
  char                pcMemory[1024*10];
  int                 nShiftPos;
  int                 boClosed;
  SendRec             sSendRec;
  int                 nReturn;
  int                 nError;
	struct sockaddr_in  sReceiveAddr;
	struct sockaddr_in  sSendAddr;
	int                 nLen;

  sSendRec = * psSendRec;

	sReceiveAddr.sin_family = AF_INET;
	sReceiveAddr.sin_port = htons (sSendRec.nReceivePort);
	sReceiveAddr.sin_addr.s_addr = inet_addr (sSendRec.szReceiveAddr);
	memset (& (sReceiveAddr.sin_zero), '\0', 8);

	sSendAddr.sin_family = AF_INET;
	sSendAddr.sin_port = htons (sSendRec.nSendPort);
	sSendAddr.sin_addr.s_addr = inet_addr (sSendRec.szSendAddr);
	memset (& (sSendAddr.sin_zero), '\0', 8);

	nStorePos = 0;

  SetScreenText ("Fully connected", 2);

  boClosed = false;
	do
	{
    // In
		if ((nMemSize - nStorePos) > 0)
		{
			// Receive some data
      nLen = sizeof (sockaddr);
  		nReturn = recvfrom (sSendRec.sReceive, (pcMemory + nStorePos), (nMemSize - nStorePos), 0, (sockaddr *)& sReceiveAddr, & nLen);
			if (nReturn < 0)
			{
			  nError = WSAGetLastError ();
			  if (nError != WSAEMSGSIZE)
			  {
  	      nReturn = 0;
  	    }
			}
			if (nReturn == 0)
			{
  			boClosed = true;
			}
			if (nReturn > 0)
			{
				nStorePos += nReturn;
			}
  	}
  	
    while ((nStorePos > 0) && (nReturn > 0))
    {
			// Send some data
      nLen = sizeof (sockaddr);
			nReturn = sendto (sSendRec.sSend, pcMemory, nStorePos, 0, (sockaddr *)& sSendAddr, nLen);
			
			if (nReturn < 0)
			{
			  nError = WSAGetLastError ();
			  if (nError != WSAEMSGSIZE)
			  {
  	      nReturn = 0;
  	    }
			}
			if (nReturn > 0)
			{
				// Shift data down
        for (nShiftPos = nReturn; nShiftPos < nStorePos; nShiftPos++)
        {
          pcMemory[nShiftPos - nReturn] = pcMemory[nShiftPos];
        }
				nStorePos -= nReturn;
			}
    }
	} while ((!boClosed) && (!gboClosed));

  SetScreenText ("Disconnected", 2);

  gboClosed = true;
	CloseSocket (sSendRec.sReceive);
	CloseSocket (sSendRec.sSend);

//  Connect (gboConnect);
}

void SetScreenText (char * szText, int nPos)
{
  WriteToLog (szText);
}

void ShowError (void)
{
  char                szError[1024];

	sprintf (szError, "Socket error (%d)", WSAGetLastError ());
  MessageBox (NULL, szError, "Socket error", MB_OK);
  SetScreenText (szError, 3);
}

void CloseSocket (SOCKET nSocket)
{
  int   nReturn;
  char  szBuf[1024];

  shutdown (nSocket, SD_SEND);

  do
  {
    nReturn = recv (nSocket, szBuf, sizeof (szBuf), 0);
  } while (nReturn > 0);

  closesocket (nSocket);
}

